/*																			*/
/* CHAREDIT.C	-- EGA character editor										*/
/*																			*/
/*	Copyright (c) 1990 Peter Anvin (all rights reserved)					*/
/*																			*/

#include "standard.h"
#include <alloc.h>
#include <graphics.h>
#include <mouse.h>
#include <bios.h>
#include <dos.h>
#include <crtplus.h>
#include <fioplus.h>
#include <miscplus.h>
#include <string.h>
#include <strplus.h>
#include <io.h>
#include <fcntl.h>
#include <sys\stat.h>
#include <conio.h>
#include <process.h>
#include <stdarg.h>

#define BGI_LINKED_IN

int capturefont(void *buffer, unsigned offset);
void loadfont(void *buffer, int height, int page);
void egaputch(int ch,int x,int y,int page,int fg,int bg,void *font,int size);
int flipbyte(int ch);
int rolb(int ch);
int rorb(int ch);
void roteightbytes(char *charblock);
void decryptscreen(void *screen, unsigned long key, unsigned bytecount);
void encryptscreen(void *screen, unsigned long key, unsigned bytecount);
extern char TESTIMAGE[25*80*2];
char RegUser[30] = "Any GPL user";
char RegID[6] = "-N/A-";

#define VERSION			"0.92"
#define COPYYEAR		"1993"

#define SCREEN_KEY		0x83F336D8UL
#define REG_KEY_0       0xC355U
#define REG_KEY_1       0x364BU

const unsigned char far *bios_lines = (char far *) 0x00400084UL;

#define bound(X,MIN,MAX)	( (X) >= (MIN) && (X) <= (MAX) )
#define MAX(A,B)			( ((A)>(B)) ? (A):(B) )
#define MIN(A,B)			( ((A)<(B)) ? (A):(B) )


#define FONT_WID		8				/* Width of font square				*/

#define BACKGND			LIGHTCYAN		/* Background colour				*/
#define DELIMITER		DARKGRAY		/* Box delimiter colour				*/

#define GRID_X(L,S)		(293 - ((S)*FONT_WID)/2)
#define GRID_Y(L,S)		5
#define GRID_COL		DARKGRAY		/* Colour of grid lines				*/
#define GRID_SQON		BLACK			/* Colour of active square			*/
#define GRID_SQOFF		WHITE			/* Colour of passive square			*/

#define CHAR_CIE		YELLOW			/* Foreground character in edit		*/
#define CHAR_MARK		LIGHTGREEN		/* Colour of marker box				*/
#define CHAR_FG			LIGHTRED		/* Foreground of character box		*/
#define CHAR_DISFG		LIGHTGRAY		/* Foreground of disabled character	*/
#define CHAR_BG			BLACK			/* Background of character box		*/
#define CHAR_X			371
#define CHAR_X1			(CHAR_X+32*FONT_WID+7)
#define CHAR_Y			5

#define TEMP_X			80				/* Temporary storage area			*/
#define TEMP_Y			100
#define TEMP_X1			(TEMP_X+16*FONT_WID+7)
#define TEMP_FIRST		512				/* First temp character number		*/
#define TEMP_CNT		32				/* Characters in temp box			*/
#define TEMP_NAME		"Temporary"		/* Name to show in monter box		*/
const unsigned char *temp_init= ("Temporary usage " "area (not saved)");
										/* Initial temp box text			*/

#define MENU_FG			BLACK			/* Menu boxes foreground			*/
#define MENU_UNAV		LIGHTGRAY		/* Unavailable foreground			*/
#define MENU_BG			WHITE			/* Menu boxes background			*/
#define FTME_X			5
#define FTME_Y			319
#define CHME_X			5
#define CHME_Y			(314-16*14)

#define MONT_X			93
#define MONT_Y			5
#define MONT_BG			WHITE			/* Foreground of "monter" box		*/
#define MONT_FG			BLACK			/* Background of "monter" box		*/

#define FINF_X			93				/* Font information box				*/
#define FINF_Y			185
#define FINF_XEXP		9				/* X position in expanded state		*/
#define FINF_BG			WHITE
#define FINF_FG         BLACK

#define FILE_X			48				/* Save/retrieve filename box		*/
#define FILE_Y			182
#define FILE_BG			WHITE
#define FILE_FG			BLACK
#define FILE_DEF		BLUE			/* Default file name foreground		*/

#define ERRM_X			80				/* Error message box				*/
#define ERRM_Y			156				/* ERRM_Y reduced by 6 per msg line	*/
#define ERRM_BG			WHITE
#define ERRM_EHDL		RED				/* ERROR/WARNING headline colour	*/
#define ERRM_IHDL		BLUE			/* INFORMATION headline colour		*/
#define ERRM_FG			BLACK

#define NAMELESS 		"Unnamed"		/* What to write if we don't know	*/
#define FILEEXT			".PSF"			/* Default file extension			*/
#define FILEID			0x0436			/* .PSF file ID						*/

#define SAVECORE		8192			/* Memory saved from UNDO malloc	*/

typedef
	unsigned char CharMatrix[32];		/* Bit matrix for one character		*/

CharMatrix char_data[512+TEMP_CNT];		/* Edit-data matrix					*/
CharMatrix far *save_data;				/* Saved-data matrix (in far core)	*/
CharMatrix save_cie;					/* save_data for current scr.cie	*/

struct scrn_status						/* Edit screen status				*/
{
	int vga;							/* True if runnning on VGA			*/
	int page;							/* Page used for screen output		*/
	int pageok;							/* OK to switch pages?				*/
	int setpal;							/* Zero if palette is not to be set	*/
	int mouspeed;						/* Mouse speed mode					*/
	int mmovx;							/* Default mouse x movement			*/
	int mmovy;							/* Default mouse y movement			*/
	int mdbl;							/* Double speed treshold			*/

	int fontsize;						/* Lines per font					*/
	int fontlen;						/* 256/512-chars					*/
	int use32;							/* Use all 32 bytes?				*/
	int group;							/* Group command mode?				*/
	int usefont;						/* Use font in Shell or exit		*/

	int cie;							/* Character in edit				*/
	int ciex, ciey;						/* Coordinates in font map			*/

	int x0, y0;							/* Upper-left of edit matrix		*/
	int x1, y1;							/* Lower-right of edit matrix		*/
	int step;							/* Size of squares in matrix		*/

	int firstch;						/* First character in current font	*/
	int fontno;							/* = 1 if 256 ch/font 1				*/
    int charfirst;						/* Font chart character start		*/
    int charcount;						/* Number of character displayed	*/
	int char_y1;						/* Lower limit of font chart		*/
    int temp_y1;						/* Ditto of temporary chart			*/

	int org_y0, org_y1;					/* Y Limits of Original box			*/

	int inedit;							/* Flag set while in the edit loop	*/

	int modif[2];						/* TRUE if modified					*/
	char *filename[2];					/* Name/ext only of saved file		*/
	char filespec[2][64];				/* Name of save file				*/
}
	scr;


enum undo_classes
{
	UND_ALWAYS, UND_BITEDIT, UND_INVERT, UND_FLIPH, UND_FLIPV,
	UND_ROTH, UND_ROTV, UND_CLEAR, UND_TURN, UND_ORIG
};

struct undo_item
{
	int chr;							/* Character stored					*/
	enum undo_classes class;			/* Class of action					*/
	CharMatrix data;					/* Character cell data				*/
};


struct undo_status
{
	int cnt;							/* Count of available UNDOs			*/
	int recnt;							/* Count of available REDOs			*/
	int free;							/* Count of free data blocks		*/
	int size;							/* cnt + recnt + free = total size	*/

	struct undo_item *ptr;				/* Pointer to last done item		*/
	struct undo_item *block;			/* Data block pointer				*/
	struct undo_item *endblock;			/* End-of-block pointer				*/
}
	undo;



void buttonwait(void)
/* Wait for a key press or a mouse button									*/
{
	int b1, b2, k;

	do
	{
		mousepress(0,&b1,NULL,NULL);
		mousepress(1,&b2,NULL,NULL);
		k = bioskey(1);
	}
	while ( b1 == 0 && b2 == 0 && k == 0 );

	if ( k )
		bioskey(0);
}


char *packfont(void *dest, void *sco, int height)
/* Converts a 32-byte-fixed font to a variable-sized font to be passed to	*/
/* BIOS or stored in a .PSF file											*/
/* Returns the first free byte after the packed font = (dest + 256*height)	*/
{
	char *inptr, *outptr;
	int i;

	inptr = sco; outptr = dest;

	for ( i = 0 ; i < 256 ; i++ )
	{
		memcpy(outptr,inptr,height);

		inptr += 32;
		outptr += height;
	}

	return(outptr);
}

int flipbyte(int ch)
/* Reverse the bits in a byte												*/
{
	int i;
	unsigned char ic = ch;
	unsigned char oc = 0;

	for ( i = 0 ; i < 8 ; i++ )
	{
		oc = (oc >> 1) | (ic & 0x80);
		ic = ic << 1;
	}

	return oc;
}

int rolb(int ch)
/* Rotates a byte left														*/
{
	unsigned char ic = ch;
	return ((ic << 1) & 0xFE) | ((ic >> 7) & 0x01);
}

int rorb(int ch)
/* Rotates a byte right														*/
{
	unsigned char ic = ch;
	return ((ic >> 1) & 0x7F) | ((ic << 7) & 0x80);
}

void roteightbytes(char *charblock)
/* Rotates an 8x8 block of bytes 90 degrees									*/
{
	unsigned char buf[8];
	unsigned char ch, bit;
	int i;
	int j;

	memset(buf, 0, 8);

	for ( i = 0 ; i < 8 ; i++ )
	{
		ch  = charblock[i];
		bit = 1 << i;
		for ( j = 0 ; j < 8 ; j++ )
		{
			buf[7-j] |= (ch & (1 << j)) ? bit : 0;
		}
	}

	memcpy(charblock, buf, 8);
}

void usefont(void *buffer, int height, int page)
/* Loads a font onto the screen, and inhibits underlining					*/
{
	void *tmpbuf;

	tmpbuf = malloc(height * 256);
	packfont(tmpbuf,buffer,height);
	loadfont(tmpbuf,height,page);
	free(tmpbuf);

	/* Just in case, no word I/O is performed								*/

    disable();
	outportb(0x3D4,0x14);				/* Underline position register		*/
	outportb(0x3D5,0x1F);				/* 32 lines below start of char		*/
    enable();
}


void defpalette(int which)
/* Sets a sharper palette than the default, or set sharp = weak colours		*/
{
    static unsigned char palette[2][17] =
	{
		{	000, 001, 002, 003, 004, 005, 024, 007,
			070, 011, 022, 033, 044, 055, 066, 077,
			000 },	/* Overscan register			*/

		{   000, 001, 002, 003, 004, 005, 024, 007,
			000, 001, 002, 003, 004, 005, 024, 007,
			000 }
	};


	if ( scr.setpal )
	{
		_DX = (unsigned) palette[which];
		_ES = _DS;
		_AX = 0x1002;
		geninterrupt(0x10);
	}
}

void refreshscreen(void);					/* Prototypes needed below		*/
void quickmouse(int speed);

int begingraph(void)
{
	int grdrvr, grmode, b4mouse, grerror;

	scr.vga = FALSE;						/* Presume no VGA				*/
	scr.pageok = TRUE;						/* Assume this					*/

#ifdef BGI_LINKED_IN
	if ( registerbgidriver(EGAVGA_driver) < 0 ) exit(255);
#endif

	detectgraph(&grdrvr,&grmode);

	switch( grdrvr )
	{
		case VGA :
					scr.vga = TRUE;			/* Set VGA flag to be true		*/

		case EGA :							/* At present, EGA64 and		*/
											/* EGAMONO not supported		*/
					grdrvr = EGA;
					grmode = EGAHI;

#ifdef BGI_LINKED_IN
					initgraph( &grdrvr, &grmode, NULL );
#else
					initgraph( &grdrvr, &grmode, getenv("BGIDIR") );
#endif
					if ( (grerror = graphresult()) != 0 )
					{
						cprintf("\r\n\nGRAPHICS INIT ERROR: %s\r\n\n",
							grapherrormsg(grerror));
						exit(1);
					}

					defpalette(0);
					settextstyle(DEFAULT_FONT,HORIZ_DIR,1);

					scr.page = 1;			/* Initial screen page 0		*/
					break;

		default :
					grdrvr = -2;			/* No EGA/VGA found				*/
					break;
	}

	if ( grdrvr >= 0 )
	{
		resetmouse();

		getmousetouch(&scr.mmovx,&scr.mmovy,&scr.mdbl);

		b4mouse = scr.mouspeed;
		scr.mouspeed = 0;
		refreshscreen();
		quickmouse(b4mouse);

		if ( _osmajor >= 10 )				/* OS/2 doesn't like the mouse	*/
		{									/* on pages > 0					*/
			scr.pageok = FALSE;
			scr.page = 0;
		}

	}

	return( (grdrvr<0) ? grdrvr : 0 );
}


void endgraph(void)
/* Finish up what's needed before terminating graphics mode					*/
{
	setmousetouch(scr.mmovx,scr.mmovy,scr.mdbl);
	closegraph();
	textmode(C80);

	resetmouse();

	if ( scr.usefont )
	{
		int selector;

		usefont(char_data[0],scr.fontsize,0);
		usefont(char_data[256],scr.fontsize,1);

		if ( scr.fontlen > 256 )
		{
			selector = 0x04;                        /* 512 ch font			*/
			defpalette(1);							/* 8-colour palette		*/
		}
		else
		{
			defpalette(0);							/* 16-colour palette	*/
			selector = scr.fontno ? 0x05 : 0x00;
		}

		_BL = selector;
		_AX = 0x1103;                               /* Select font pages	*/
		geninterrupt(0x10);
    }

	_video.screenheight = *bios_lines + 1;
	window(1,1,80,_video.screenheight);

	textattr(LIGHTGRAY);
	clrscr();
}


void drawrel(int *dxylist)
{
	int op, dx, dy;

	while( ( op = *(dxylist++) ) >= 0 )
	{
		dx = *(dxylist++);
		dy = *(dxylist++);

		if ( op )
			linerel(dx,dy);
		else
			moverel(dx,dy);
	}
}


enum err_types
{
	ERR_FILE, ERR_PROG, ERR_EXT, ERR_INFO, ERR_DISAB, ERR_WARN,
    ERR_NOTDONEYET
};

int errormsg(enum err_types type,...)
{
/*																			*/
/* The first letter in the headline[] represents the message type			*/
/* 'e' = ERROR: Display error headline, no return							*/
/* 'i' = INFORMATION: Display information headline, no return				*/
/* 'w' = WARNING: Display error headline, select PROCEED (1) or CANCEL (0)	*/
/*																			*/
/* The second character is the number of information lines to print.  All 	*/
/* lines are expected to be supplied.										*/
/*																			*/
	static char *headline[] =
	{
		"e\2FILE I/O ERROR",
		"e\2PROGRAM ERROR - NOTIFY AUTHOR",
		"e\2EXTERNAL ERROR",
		"i\3PROGRAM INFORMATION",
		"e\2DISABLED COMMAND",
		"w\2DATA LOSS WARNING",
        "e\2UNIMPLEMENTED COMMAND - NOTIFY AUTHOR"

	};

	int ret = 0;				/* If nothing to return, return zero		*/
	int dy, i;
	int c, x, y;				/* Mouse press count/coordinates			*/
	va_list args;				/* Argument list type						*/

	va_start(args,type);

	dy = 6 * headline[type][1];		/* Y offset up/down						*/

	setcolor(DELIMITER);
	setfillstyle(SOLID_FILL,ERRM_BG);
	bar3d(ERRM_X,ERRM_Y-dy,ERRM_X+480,ERRM_Y+50+dy,0,0);

	settextjustify(CENTER_TEXT,TOP_TEXT);

	setcolor( (headline[type][0] == 'i') ? ERRM_IHDL : ERRM_EHDL );

	outtextxy(ERRM_X+240,ERRM_Y+3-dy,headline[type]+2);

	setcolor(ERRM_FG);
	y = ERRM_Y+25-dy;

	for ( i = headline[type][1] ; i ; i-- )
	{
		outtextxy(ERRM_X+240,y,va_arg(args,char *));
		y += 12;
	}

	if ( headline[type][0] != 'w' )
	{
		outtextxy(ERRM_X+240,ERRM_Y+40+dy,"Press any key or mouse button to continue");
		buttonwait();
	}
	else
	{
		outtextxy(ERRM_X+169,ERRM_Y+40+dy,"Please select Proceed or Cancel.");

		setcolor(DELIMITER);
		rectangle(ERRM_X+337,ERRM_Y+34+dy,ERRM_X+407,ERRM_Y+48+dy);
		rectangle(ERRM_X+407,ERRM_Y+34+dy,ERRM_X+477,ERRM_Y+48+dy);

		setcolor(MENU_FG);
		outtextxy(ERRM_X+372,ERRM_Y+38+dy,"Proceed");
		outtextxy(ERRM_X+442,ERRM_Y+38+dy,"Cancel");

		mouseon();

		ret = -1;

		do
		{
			mousepress(0,&c,&x,&y);

			if ( c != 0 )
			{
				y -= dy;

				if ( bound(x,ERRM_X+338,ERRM_X+476) &&
					 bound(y,ERRM_Y+35,ERRM_Y+47) &&
					 getpixel(x,y) != DELIMITER )
				{
					ret = ( x < ERRM_X+407 );
				}
			}
			else if ( bioskey(1) )
			{
				c = bioskey(0);
				if ( c & 0xFF00 )
				{
					c &= 0x00FF;
					if ( c == 'P' || c == 'p' )
						ret = 1;
					else if ( c == 'C' || c == 'c' )
						ret = 0;
				}
			}
		}
		while ( ret == -1 );

		mouseoff();
	}

	va_end(args);

	return(ret);
}



void selectchar(int ch)
{
	int x, y, i, m;
	char *data, *name, cienum[14];

	static int brace[] =
		{	0,-2,-2, 1,-1,0, 1,0,-1, 1,13,0, 1,0,1, 1,-1,0,
			0,0,0, 1,1,0, 1,0,1, 1,-13,0, 1,0,-1, 1,1,0, -1 };

	brace[20] = 3+scr.fontsize;

	if ( scr.cie != ch )
	{
    	if ( scr.ciex )
        {
			setcolor(CHAR_BG);
			moveto(scr.ciex,scr.ciey);
			drawrel(brace);
			egaputch(scr.cie,scr.ciex,scr.ciey,scr.page,bound(scr.cie,scr.firstch,scr.firstch+scr.fontlen-1)?CHAR_FG:CHAR_DISFG,CHAR_BG,char_data,scr.fontsize);
        }

		if ( scr.cie/scr.fontlen != ch/scr.fontlen )
        {
        	setfillstyle(SOLID_FILL,MONT_BG);
            bar(MONT_X+1,MONT_Y+1,MONT_X+111,MONT_Y+9);

	        setcolor(MONT_FG);

	        settextjustify(CENTER_TEXT,TOP_TEXT);
            if ( ch >= TEMP_FIRST )
            {
            	name = TEMP_NAME;
            }
            else
            {
            	name = ( scr.filename[ch/scr.fontlen] == NULL ) ? NAMELESS : scr.filename[ch/scr.fontlen];
            }

	        outtextxy(MONT_X+56,MONT_Y+2, name );
        }

        scr.cie = ch;
	}

	if ( bound(scr.cie,scr.charfirst,scr.charfirst+scr.charcount-1) )
	{
		scr.ciex = CHAR_X + 4 + FONT_WID*(scr.cie & 0x1F);
		scr.ciey = CHAR_Y + 4 + (scr.fontsize+4)*((scr.cie-scr.charfirst) >> 5);
	}
	else if ( scr.cie >= TEMP_FIRST )
    {
    	scr.ciex = TEMP_X + 4 + FONT_WID*(scr.cie & 0x0F);
		scr.ciey = TEMP_Y + 4 + (scr.fontsize+4)*((scr.cie-TEMP_FIRST) >> 4);
    }
	else
		scr.ciex = 0;

	if ( scr.ciex )
    {
		egaputch(scr.cie,scr.ciex,scr.ciey,scr.page,CHAR_CIE,CHAR_BG,char_data,scr.fontsize);
		setcolor(CHAR_MARK);
		moveto(scr.ciex,scr.ciey);
		drawrel(brace);
    }

	farmemcpy(save_cie,save_data[scr.cie],32);

	data = char_data[scr.cie];

	y = scr.y0 + 1;

	for ( i = 0 ; i < scr.fontsize ; i++ )
	{
		x = scr.x0 + 1;

		for ( m = 0x80 ; m ; m >>= 1 )
		{
			setfillstyle( SOLID_FILL, (*data & m) ? GRID_SQON : GRID_SQOFF );
			bar(x,y,x+scr.step-2,y+scr.step-2);

			x += scr.step;
		}

		y += scr.step;
		data++;
	}

	setfillstyle(SOLID_FILL,MONT_BG);
	setcolor(DELIMITER);
	bar(MONT_X+1,MONT_Y+11,MONT_X+99-FONT_WID,scr.org_y0-1);
	egaputch(scr.cie,MONT_X+107-FONT_WID,MONT_Y+13,scr.page,MONT_FG,MONT_BG,char_data,scr.fontsize);
	egaputch(0,MONT_X+107-FONT_WID,scr.org_y0+3,scr.page,MONT_FG,MONT_BG,save_cie,scr.fontsize);

    if ( scr.cie >= TEMP_FIRST )
    {
    	sprintf(cienum,"Slot %02d",scr.cie-TEMP_FIRST+1);
    }
    else
    {
		sprintf(cienum,"%3dd %03Xh",scr.cie % scr.fontlen, scr.cie % scr.fontlen);
    }
#if MONT_FG != DELIMITER
	setcolor(MONT_FG);
#endif
	settextjustify(CENTER_TEXT,CENTER_TEXT);

	outtextxy(MONT_X+46,MONT_Y+13+MAX(scr.fontsize,8)/2,cienum);
}


void invertmenubox(int x, int y)
/* Reverses the colour on a standard-sized (70x14) menu box					*/
{
	int i;

	setwritemode(1);						/* XOR operation				*/
	setcolor(MENU_FG^MENU_BG);

	for ( i = y+1 ; i < y+14 ; i++ )
	{
		line(x+1,i,x+69,i);
	}

	setwritemode(0);						/* Reset to WRITEOVER			*/
}


void quickmouse(int speed)
{
	int dx, dy;

	invertmenubox(FTME_X+140-70*scr.mouspeed,FTME_Y+14);
	scr.mouspeed = speed;

	dx = scr.mmovx + ( ( speed * scr.mmovx ) >> 1 );
	dy = scr.mmovy + ( ( speed * scr.mmovy ) >> 1 );

	setmousetouch(dx,dy,scr.mdbl);

	invertmenubox(FTME_X+140-70*scr.mouspeed,FTME_Y+14);
}


void refreshscreen(void)
{
	static int grmouse[] =
	{
		1,0,2, 1,-5,0, 1,0,8, 1,1,0, 1,0,-8,
		0,2,2, 1,1,0, 1,0,2, 1,-1,0, 1,0,-1,
		0,3,-1, 1,1,0, 1,0,2, 1,-1,0, 1,0,-1,
		0,-4,5, 1,7,0, 1,0,-8, 1,1,0, 1,0,8,
		0,-5,-8, 1,3,0, -1
	};

	int x, y, xc, yc, yn, i;
	char text[14];

	scr.page ^= scr.pageok;

	setactivepage(scr.page);

	setfillstyle(SOLID_FILL,BACKGND);		/* Background screen			*/
	bar(0,0,639,349);						/* Fill the screen				*/


	/* ---- Draw edit grid ----												*/

/*	i = ( scr.fontsize < 32 ) ? scr.fontsize+1 : scr.fontsize; */
	i = scr.fontsize;
	scr.step = MIN( 307/i, 18 );

	scr.x0 = GRID_X(scr.fontsize,scr.step);
	scr.x1 = scr.x0 + scr.step * FONT_WID;
	scr.y0 = GRID_Y(scr.fontsize,scr.step);
	scr.y1 = scr.y0 + scr.step * scr.fontsize;

	setfillstyle(SOLID_FILL,GRID_SQOFF);
	bar(scr.x0,scr.y0,scr.x1,scr.y1);

	setcolor(GRID_COL);

	x = FONT_WID * scr.step;
	y = scr.y0;

	for ( i = 0 ; i <= scr.fontsize ; i++ )
	{
		moveto( scr.x0, y );
 		linerel( x, 0 );

		y += scr.step;
	}

	x = scr.x0;
	y = scr.fontsize * scr.step;

	for ( i = 0 ; i <= FONT_WID ; i++ )
	{
		moveto( x, scr.y0 );
		linerel( 0, y );

		x += scr.step;
	}


	/* ---- Draw "monter" box												*/

	scr.fontno = ( scr.fontlen > 256 ) ? 0 : scr.firstch >> 8 ;

	scr.org_y0 = MONT_Y+14+MAX(scr.fontsize,8);
	scr.org_y1 = MONT_Y+18+2*MAX(scr.fontsize,8);

	setfillstyle(SOLID_FILL,MONT_BG);
	setcolor(DELIMITER);
	bar3d(MONT_X,MONT_Y,MONT_X+112,scr.org_y1,0,0);
	line(MONT_X,MONT_Y+10,MONT_X+112,MONT_Y+10);
	line(MONT_X+100-FONT_WID,MONT_Y+10,MONT_X+100-FONT_WID,scr.org_y1);
	line(MONT_X,scr.org_y0,MONT_X+112,scr.org_y0);

	setcolor(MONT_FG);

	settextjustify(CENTER_TEXT,TOP_TEXT);
	outtextxy(MONT_X+56,MONT_Y+2, ( scr.filename[scr.cie/scr.fontlen] == NULL ) ? NAMELESS : scr.filename[scr.cie/scr.fontlen] );

	settextjustify(CENTER_TEXT,CENTER_TEXT);
	outtextxy(MONT_X+46,scr.org_y0+3+MAX(scr.fontsize,8)/2,"Original");


	/* ---- Draw font info box ----											*/

	setfillstyle(SOLID_FILL,FINF_BG);
	setcolor(DELIMITER);
	bar3d(FINF_X,FINF_Y,FINF_X+112,FINF_Y+30,0,0);

	setcolor(FINF_FG);

	settextjustify(CENTER_TEXT,TOP_TEXT);

	sprintf(text,"Point size%3d",scr.fontsize);
	outtextxy(FINF_X+56,FINF_Y+2,text);

	sprintf(text,"EGA lines%4d",350/scr.fontsize);
	outtextxy(FINF_X+56,FINF_Y+12,text);

	sprintf(text,"VGA lines%4d",400/scr.fontsize);
	outtextxy(FINF_X+56,FINF_Y+22,text);


	/* ---- Draw font chart box ----										*/

	setfillstyle(SOLID_FILL,CHAR_BG);		/* Character box				*/
	setcolor(DELIMITER);

    yn = MIN(16,303/(scr.fontsize+4));
    scr.charcount = yn*32;

	bar3d(CHAR_X,CHAR_Y,CHAR_X+32*FONT_WID+8,CHAR_Y+yn*(scr.fontsize+4)+4,0,0);

    if ( scr.charfirst < 0 )
    	scr.charfirst = 0;

    if ( scr.charfirst + scr.charcount > 512 )
    	scr.charfirst = 512 - scr.charcount;

    i = scr.charfirst;
	yc = CHAR_Y + 4;

	for ( y = 0 ; y < yn ; y++ )
	{
		xc = CHAR_X + 4;

		for ( x = 0 ; x < 32 ; x++ )
		{
			egaputch(i++,xc,yc,scr.page,bound(i,scr.firstch,scr.firstch+scr.fontlen-1)?CHAR_FG:CHAR_DISFG,CHAR_BG,char_data,scr.fontsize);
			xc += FONT_WID;
		}

		yc += scr.fontsize + 4;
	}

	scr.char_y1 = yc-1;

    /* ---- Draw temp storage box ----										*/

    bar3d(TEMP_X,TEMP_Y,TEMP_X+16*FONT_WID+8,TEMP_Y+2*scr.fontsize+12,0,0);

	i = TEMP_FIRST;
    yc = TEMP_Y + 4;

    for ( y = 0 ; y < 2 ; y++ )
    {
    	xc = TEMP_X + 4;

        for ( x = 0 ; x < 16 ; x++ )
        {
        	egaputch(i++,xc,yc,scr.page,CHAR_DISFG,CHAR_BG,char_data,scr.fontsize);
            xc += FONT_WID;
        }

        yc += scr.fontsize + 4;
    }

    scr.temp_y1 = yc-1;

	/* ---- Draw font menu at bottom of screen -----						*/

	setfillstyle(SOLID_FILL,MENU_BG);
	setcolor(DELIMITER);
	settextjustify(CENTER_TEXT,CENTER_TEXT);

	for ( i = FTME_X ; i < FTME_X + 9*70 ; i += 70 )
		bar3d(i,FTME_Y,i+70,FTME_Y+28,0,0);

	line(FTME_X,FTME_Y+14,FTME_X+9*70,FTME_Y+14);

#if DELIMITER != MENU_FG
	setcolor(MENU_FG);
#endif
	outtextxy(FTME_X+175,FTME_Y+8,"Save");
	outtextxy(FTME_X+245,FTME_Y+8,"Retrieve");
	outtextxy(FTME_X+455,FTME_Y+8,"256 ch");
	outtextxy(FTME_X+525,FTME_Y+8,"512 ch");
	outtextxy(FTME_X+42,FTME_Y+22,"fast");
	moveto(FTME_X+14,FTME_Y+16);
	drawrel(grmouse);
	outtextxy(FTME_X+113,FTME_Y+22,"med");
	moveto(FTME_X+89,FTME_Y+16);
	drawrel(grmouse);
	outtextxy(FTME_X+182,FTME_Y+22,"slow");
	moveto(FTME_X+154,FTME_Y+16);
	drawrel(grmouse);

	outtextxy(FTME_X+315,FTME_Y+22,"Test");
	outtextxy(FTME_X+455,FTME_Y+22,"Info");
	outtextxy(FTME_X+525,FTME_Y+22,"OS Shell");
	outtextxy(FTME_X+595,FTME_Y+22,"Exit");
	outtextxy(FTME_X+385,FTME_Y+22,"Use Font");


	setcolor(
		( scr.fontlen == 256 ) &&
		( scr.fontsize == 8 || scr.fontsize == 14 || ( scr.vga && scr.fontsize == 16 ) )
		? MENU_FG : MENU_UNAV
		);
	outtextxy(FTME_X+245,FTME_Y+22,"Copy ROM");

	setcolor( ( scr.fontsize > 2 ) ? MENU_FG : MENU_UNAV );
	outtextxy(FTME_X+35,FTME_Y+8,"Shrink");
	setcolor( ( scr.fontsize < 32 ) ? MENU_FG : MENU_UNAV );
	outtextxy(FTME_X+105,FTME_Y+8,"Expand");

	invertmenubox( FTME_X+140-70*scr.mouspeed, FTME_Y+14 );

	setcolor( *RegID ? MENU_FG : MENU_UNAV );
	outtextxy(FTME_X+595,FTME_Y+8,"Group");

	if ( scr.group )
		invertmenubox(FTME_X+560,FTME_Y);

	invertmenubox( ( scr.fontlen <= 256 ) ? FTME_X+420 : FTME_X+490, FTME_Y );

	if ( scr.usefont )
		invertmenubox(FTME_X+350,FTME_Y+14);

	if ( scr.fontlen <= 256 )
	{
		setcolor(MENU_FG);
		outtextxy(FTME_X+315,FTME_Y+8,"Font 0");
		outtextxy(FTME_X+385,FTME_Y+8,"Font 1");
		invertmenubox( scr.fontno ? FTME_X+350 : FTME_X+280, FTME_Y );
	}
	else
	{
		bar(FTME_X+281,FTME_Y+1,FTME_X+419,FTME_Y+13);

		setcolor(DELIMITER);
		for ( i = FTME_X+339 ; i < FTME_X+339+9*9 ; i += 9 )
			line(i,FTME_Y,i,FTME_Y+14);

		setcolor(MENU_FG);
		sprintf( text, "%d-%d", scr.charfirst, scr.charfirst+scr.charcount-1 );
		outtextxy(FTME_X+309,FTME_Y+8,text);

		setfillstyle(SOLID_FILL,MENU_FG);

        for ( xc = 0 ; xc <= 256 ; xc += 32 )
        {
        	if ( xc >= scr.charfirst && xc+255 <= scr.charfirst+scr.charcount )
            {
				i = FTME_X + 340 + (xc >> 5) * 9;
				bar(i,FTME_Y+1,i+7,FTME_Y+13);
			}
        }

		setfillstyle(SOLID_FILL,MENU_BG);
	}

	/* ---- Draw character-in-edit menu at side of screen ----				*/

	setcolor(DELIMITER);

	for ( i = CHME_Y ; i < CHME_Y + 16*14 ; i += 14 )
		bar3d(CHME_X,i,CHME_X+70,i+14,0,0);

#if DELIMITER != MENU_FG
	setcolor(MENU_FG);
#endif
	outtextxy(CHME_X+35,CHME_Y+8,"Invert");
	outtextxy(CHME_X+35,CHME_Y+22,"Flip \x1D");		/* 0x1D = Right/left	*/
	outtextxy(CHME_X+35,CHME_Y+36,"Flip \x12");		/* 0x12 = Up/down		*/
	outtextxy(CHME_X+35,CHME_Y+50,"Rotate \x18");	/* 0x18 = Up arrow		*/
	outtextxy(CHME_X+35,CHME_Y+64,"Rotate \x19");	/* 0x19 = Down arrow	*/
	outtextxy(CHME_X+35,CHME_Y+78,"Rotate \x1B");	/* 0x1B = Left arrow	*/
	outtextxy(CHME_X+35,CHME_Y+92,"Rotate \x1A");	/* 0x1C = Right arrow	*/
	outtextxy(CHME_X+35,CHME_Y+120,"Del Row");
	outtextxy(CHME_X+35,CHME_Y+134,"Ins Row");
	outtextxy(CHME_X+35,CHME_Y+148,"Del Col");
	outtextxy(CHME_X+35,CHME_Y+162,"Ins Col");
	outtextxy(CHME_X+35,CHME_Y+176,"Clear");

	if ( scr.fontsize != 8 )
		setcolor(MENU_UNAV);

	outtextxy(CHME_X+35,CHME_Y+106,"Turn 90");

	setcolor( undo.cnt ? MENU_FG : MENU_UNAV );
	outtextxy(CHME_X+35,CHME_Y+190,"Undo");

	setcolor( undo.recnt ? MENU_FG : MENU_UNAV );
	outtextxy(CHME_X+35,CHME_Y+204,"Redo");

	setcolor( ( scr.fontsize == 32 ) ? MENU_UNAV : MENU_FG );
	outtextxy(CHME_X+35,CHME_Y+218,"32 Rows");
	if ( scr.use32 )
		invertmenubox(CHME_X,CHME_Y+210);

	/* ---- Update character-specific screen info ----						*/

	selectchar(scr.cie);

	setvisualpage(scr.page);
	setmousepage(scr.page);
}


void registerstep(enum undo_classes class)
/* Register the step just taken into the UNDO buffer, if applicable.		*/
{
	scr.modif[scr.fontno] = TRUE;

	if ( class == UND_ALWAYS || undo.cnt == 0 || (undo.ptr)->chr != scr.cie || (undo.ptr)->class != class )
	{
		if ( undo.recnt )
		{
			undo.free += undo.recnt;
			undo.recnt = 0;
			setcolor( MENU_UNAV );
			outtextxy(CHME_X+35,CHME_Y+204,"Redo");
		}

		if ( undo.cnt == 0 )
		{
			setcolor( MENU_FG );
			outtextxy(CHME_X+35,CHME_Y+190,"Undo");
		}

		if ( undo.free )
		{
			undo.free--;
			undo.cnt++;
		}

		undo.ptr++;
		if ( undo.ptr >= undo.endblock )
			undo.ptr -= undo.size;

		(undo.ptr) -> chr = scr.cie;
		(undo.ptr) -> class = class;

		memcpy( (undo.ptr)->data, char_data[scr.cie], 32 );
	}
}


int selectpixelxy(int *x, int *y, int rows)
/* Selects a row/column in the character grid for insert/delete				*/
/* (rows) may or may not equal scr.fontsize									*/
/* Returns TRUE if in zone, FALSE if outside = Cancel Command				*/
{
	int mx, my, c, endgrid;

	endgrid = scr.y0 + scr.step * rows - 1;		/* We cannot rely on scr.y1	*/

	mouseon();

	while (	mousepress(0,&c,&mx,&my), c == 0 );

	if ( bound(mx,scr.x0,scr.x1-1) && bound(my,scr.y0,endgrid) )
	{
		if ( x != NULL )
			*x = ( mx - scr.x0 ) / scr.step;

		if ( y != NULL )
			*y = ( my - scr.y0 ) / scr.step;

		c = TRUE;
	}
	else
		c = FALSE;

	mouseoff();

	return(c);
}


void character_command(int cmd)
/* Command from the vertical character-in-edit command box					*/
{
	int i, j, k, fsz;
	unsigned char *chdata;

	chdata = char_data[scr.cie];
	fsz = scr.use32 ? 32 : scr.fontsize;

	switch(cmd)
	{
		case 0:										/* Command: Invert		*/
				registerstep(UND_INVERT);

				for ( i = 0 ; i < fsz ; i++ )
				{
					chdata[i] ^= 0xFF;
				}
				break;

		case 1:										/* Command: Flip H		*/
				registerstep(UND_FLIPH);

				for ( i = 0 ; i < fsz ; i++ )
				{
					chdata[i] = flipbyte( chdata[i] );
				}
				break;

		case 2: 									/* Command: Flip V		*/
				registerstep(UND_FLIPV);

				i = fsz/2;
				j = fsz - (i--);

				while ( i >= 0 )
				{
					k = chdata[i];
					chdata[i] = chdata[j];
					chdata[j] = k;

					i--;
					j++;
				}
				break;

		case 3:										/* Command: Rotate Up	*/
				registerstep(UND_ROTV);

				k = chdata[0];

				for ( i = 1 ; i < fsz ; i++ )
				{
					chdata[i-1] = chdata[i];
				}

				chdata[fsz-1] = k;

				break;

		case 4:										/* Command: Rotate Dn	*/
				registerstep(UND_ROTV);

				k = chdata[fsz-1];

				for ( i = fsz-1 ; i > 0 ; i-- )
				{
					chdata[i] = chdata[i-1];
				}

				chdata[0] = k;

				break;

		case 5:                   					/* Command: Rotate Lft	*/
				registerstep(UND_ROTH);

				for ( i = 0 ; i < fsz ; i++ )
				{
					chdata[i] = rolb( chdata[i] );
				}
				break;

		case 6:                 					/* Command: Rotate Rght	*/
				registerstep(UND_ROTH);

				for ( i = 0 ; i < fsz ; i++ )
				{
					chdata[i] = rorb( chdata[i] );
				}
				break;

		case 7:										/* Command: Turn 90	*/
				if ( scr.fontsize == 8 )
				{
					registerstep(UND_TURN);

					roteightbytes(chdata);
				}
				else
				{
					errormsg(ERR_DISAB,"ONLY SQUARE CHARACTERS CAN BE ROTATED","You can only rotate characters in 8-point fonts.");
					refreshscreen();
				}
				break;

		case 8:										/* Command: Del Row		*/
				invertmenubox(CHME_X,CHME_Y+112);

				if ( selectpixelxy(NULL,&k,scr.fontsize) )
				{
					registerstep(UND_ALWAYS);

                    for ( i = k+1 ; i < fsz ; i++ )
					{
						chdata[i-1] = chdata[i];
					}

					chdata[fsz-1] = 0;
                }
				invertmenubox(CHME_X,CHME_Y+112);

				break;

		case 9:										/* Command: Ins Row		*/
				invertmenubox(CHME_X,CHME_Y+126);

				if( selectpixelxy(NULL,&k,scr.fontsize) )
				{
					registerstep(UND_ALWAYS);

					for ( i = fsz ; i > k ; i-- )
					{
						chdata[i] = chdata[i-1];
					}

					chdata[k] = 0;
				}
				invertmenubox(CHME_X,CHME_Y+126);

				break;

		case 10:									/* Command: Del Col		*/
				{
					unsigned km, sm;

					invertmenubox(CHME_X,CHME_Y+140);

					if( selectpixelxy(&k,NULL,scr.fontsize) )
					{
						registerstep(UND_ALWAYS);

                        km = 0xFF00 >> k;			/* Bit 0 never 1		*/
						sm = 0xFE ^ km;

						for ( i = 0 ; i < 32 ; i++ )
						{
							chdata[i] = (chdata[i] & km) | ( rolb( chdata[i] ) & sm );
						}
					}

					invertmenubox(CHME_X,CHME_Y+140);
				}
				break;

		case 11:									/* Command: Ins Col		*/
				{
					unsigned km, sm;

					invertmenubox(CHME_X,CHME_Y+154);

					if( selectpixelxy(&k,NULL,scr.fontsize) )
					{
						registerstep(UND_ALWAYS);

						km = 0xFF00 >> k;
						sm = 0x7F >> k;

						for ( i = 0 ; i < 32 ; i++ )
						{
							chdata[i] = (chdata[i] & km) | ( rorb( chdata[i] ) & sm );
						}
					}

					invertmenubox(CHME_X,CHME_Y+154);
				}
				break;

		case 12:                    				/* Command: Clear		*/
				registerstep(UND_CLEAR);

				memset(chdata,0,32);
				break;

		case 13:									/* Command: Undo		*/
				if ( undo.cnt )
				{
					{
						struct undo_item *uptr;

						uptr = undo.ptr+1;
						if ( uptr >= undo.endblock )
							uptr -= undo.size;

						uptr->class = UND_ALWAYS;
						uptr->chr = scr.cie;

						memcpy( uptr->data, char_data[scr.cie], 32 );

						setcolor( MENU_FG );
						outtextxy(CHME_X+35,CHME_Y+204,"Redo");
					}

                    if ( (undo.ptr)->chr != scr.cie )
					{
						i = (undo.ptr)->chr;

						if ( bound(i,scr.firstch,scr.firstch+255) )
						{
							selectchar(i);
						}
						else
						{
							scr.firstch = i & 0xFF00;
							scr.cie = i;
							refreshscreen();
						}
					}
					else
					{
						memcpy( char_data[scr.cie], (undo.ptr)->data, 32 );
						undo.cnt--;
						undo.ptr--;

						if ( undo.ptr < undo.block )
							undo.ptr += ( undo.endblock - undo.block );

						if ( undo.cnt == 0 )
						{
							setcolor( MENU_UNAV );
							outtextxy(CHME_X+35,CHME_Y+190,"Undo");
						}

						undo.recnt++;

						selectchar(scr.cie);
					}
				}
				else
				{
					errormsg(ERR_DISAB,"UNDO BUFFER EMPTY","You must perform a command before you can Undo.");
					refreshscreen();
				}
				break;

		case 14:									/* Command: Redo		*/
				if ( undo.recnt )
				{
					struct undo_item *uptr;

					uptr = undo.ptr+2;
					if ( uptr >= undo.endblock )
						uptr -= undo.size;

					if ( uptr->chr != scr.cie )
					{
						i = uptr->chr;

						if ( bound(i,scr.firstch,scr.firstch+255) )
						{
							selectchar(i);
						}
						else
						{
							scr.firstch = i & 0xFF00;
							scr.cie = i;
							refreshscreen();
						}
					}
					else
					{
						if ( undo.cnt == 0 )
						{
							setcolor( MENU_FG );
							outtextxy(CHME_X+35,CHME_Y+190,"Undo");
						}

						undo.ptr++;
						if ( undo.ptr >= undo.endblock )
							undo.ptr -= undo.size;

						memcpy( (undo.ptr)->data, char_data[scr.cie], 32 );
						(undo.ptr)->chr = scr.cie;
						(undo.ptr)->class = UND_ALWAYS;

						memcpy( char_data[scr.cie], uptr->data, 32 );
						undo.cnt++;
						undo.recnt--;

						if ( undo.recnt == 0 )
						{
							setcolor( MENU_UNAV );
							outtextxy(CHME_X+35,CHME_Y+204,"Redo");
						}

                        selectchar(scr.cie);
					}
				}
				else
				{
					errormsg(ERR_DISAB,"REDO BUFFER EMPTY","You must Undo a command before you can Redo.");
					refreshscreen();
				}
				break;

		case 15:                                    /* Command: Use32		*/
				scr.use32 = !scr.use32;
				invertmenubox(CHME_X,CHME_Y+210);
				break;
	}

	selectchar(scr.cie);
}



char *expandedname(char *inspec)
{
	static char fullname[64];
	char partname[64];

	strcpy(partname,inspec);

	if ( strchr(partname,'.') == NULL )
		strcat(partname,FILEEXT);

	strxupr(partname);
	fqfn(partname,fullname);

	return(fullname);

}



int getfilename(char *filename, char *operation)
{
	char inspec[64];
	int n, ch, def;

	setcolor(DELIMITER);
	setfillstyle(SOLID_FILL,FILE_BG);

	bar3d(FILE_X,FILE_Y,FILE_X+544,FILE_Y+40,0,0);
	bar3d(FILE_X+17,FILE_Y+28,FILE_X+527,FILE_Y+38,0,0);

	settextjustify(CENTER_TEXT,TOP_TEXT);
	setcolor(FILE_FG);

	outtextxy(FILE_X+272,FILE_Y+5,operation);
	outtextxy(FILE_X+272,FILE_Y+15,"Please enter file name");

	n = 0;
	inspec[0] = '\0';
	settextjustify(LEFT_TEXT,TOP_TEXT);

	if ( scr.filename[scr.fontno] != NULL )
	{
#if FILE_FG != FILE_DEF
		setcolor(FILE_DEF);
#endif
		outtextxy(FILE_X+20,FILE_Y+30,scr.filespec[scr.fontno]);

#if FILE_FG != FILE_DEF
		setcolor(FILE_FG);
#endif

		def = TRUE;
	}
	else
		def = FALSE;

	do
	{
		moveto(FILE_X+20+8*n,FILE_Y+30);
		linerel(0,7);

		ch = inkey();

		setcolor(FILE_BG);
		linerel(0,-7);
		setcolor(FILE_FG);

		if ( ch >= ' ' )
		{
			if ( def )
			{
				bar(FILE_X+18,FILE_Y+29,FILE_X+526,FILE_Y+37);
				def = FALSE;
			}

			if ( n < 59 || ( n < 63 && strchr (inspec,'.') != NULL ) )
			inspec[n] = ch;
			inspec[n+1] = '\0';
			outtextxy(FILE_X+20+8*n,FILE_Y+30,&inspec[n]);
			n++;
		}
		else
		{
			switch(ch)
			{
				case KEY_BS :
								if ( def )
								{
									strcpy(inspec,scr.filespec[scr.fontno]);
									outtextxy(FILE_X+20,FILE_Y+30,inspec);

									n = strlen(inspec);
									def = FALSE;
								}

								if ( n > 0 )
								{
									n--;
									setcolor(FILE_BG);
									outtextxy(FILE_X+20+8*n,FILE_Y+30,&inspec[n]);
									setcolor(FILE_FG);
									inspec[n] = '\0';
								}
								break;

				case KEY_CR :
								if ( def )
								{
									strcpy(inspec,scr.filespec[scr.fontno]);
									outtextxy(FILE_X+20,FILE_Y+30,inspec);

									n = strlen(inspec);
									def = FALSE;
								}

								if ( n != 0 )
								{
									outtextxy(FILE_X+20,FILE_Y+30,inspec);
									ch = 9999;			/* End edit			*/
								}
								break;

				case KEY_ESC :
								ch = 9998;				/* Abort edit		*/
								break;
			}
		}
	}
	while ( ch < 9000 );

	if ( ch == 9998 )
		return(FALSE);

	strcpy(filename,expandedname(inspec));

	bar(FILE_X+18,FILE_Y+29,FILE_X+526,FILE_Y+37);
	outtextxy(FILE_X+20,FILE_Y+30,filename);

	return(TRUE);
}



void getfont(char *filspec)
{
	unsigned mode;
	int file, ch;

    if ( scr.modif[scr.fontno] )
    {
    	if ( !errormsg(ERR_WARN,"THE CURRENT FONT HAS NOT BEEN SAVED","Select Cancel to save this font before loading another.") )
        	return;
    }

	file = open( filspec, O_RDONLY | O_EXCL | O_BINARY );

	if ( file == -1 )
	{
		errormsg(ERR_FILE,"CANNOT OPEN FILE","The file is nonexistent or the filename was wrong.");
		return;
	}

	read(file,&mode,2);             /* Get file ID							*/

	if ( mode != FILEID )
	{
		close(file);
		errormsg(ERR_FILE,"UNKNOWN FILE FORMAT","The file is not a valid PC Screen Font file.");
		return;
	}

	mode = 0;
	read(file,&mode,1);				/* Get file mode						*/

	if ( mode > 1 )
	{
		close(file);
		errormsg(ERR_FILE,"UNKNOWN FILE MODE NUMBER","You need a newer version of CHAREDIT.");
		return;
	}

	if ( mode == 0 )
	{
		if ( scr.fontlen > 256 )
		{
			close(file);
			errormsg(ERR_FILE,"256 CHARACTER FONT","Load in 256 character mode.");
			return;
		}

		memset(char_data[scr.firstch],0,32*256);

		scr.fontsize = 0;
		read(file,&scr.fontsize,1);	/* Read character height				*/

		for ( ch = scr.firstch ; ch < scr.firstch+256 ; ch++ )
		{
			read(file,char_data[ch],scr.fontsize);
		}

		close(file);

		farmemcpy(save_data[scr.firstch],char_data[scr.firstch],32*256);
	}
	else
	{
		if ( scr.fontlen <= 256 )
		{
			close(file);
			errormsg(ERR_FILE,"512 CHARACTER FONT","Load in 512 character mode.");
			return;
		}

		memset(char_data[0],0,32*512);

		scr.fontsize = 0;
		read(file,&scr.fontsize,1);	/* Read character height				*/

		for ( ch = 0 ; ch < 512 ; ch++ )
		{
			read(file,char_data[ch],scr.fontsize);
		}

		close(file);

		farmemcpy(save_data[0],char_data[0],32*512);
	}

	undo.cnt = undo.recnt = 0;
	undo.free = undo.size;
	undo.ptr = undo.block;
	undo.endblock = undo.block + undo.size;

	scr.modif[scr.fontno] = FALSE;

	fqfn(filspec,scr.filespec[scr.fontno]);
	strxupr(scr.filespec[scr.fontno]);

	scr.filename[scr.fontno] = strrchr(scr.filespec[scr.fontno],'\\') + 1;
}


void savefont(char *filspec)
{
	unsigned mode;
	int file, ch;

	file = open( filspec, O_WRONLY | O_CREAT | O_TRUNC | O_BINARY, S_IWRITE );

	if ( file == -1 )
	{
		errormsg(ERR_FILE,"CANNOT OPEN/CREATE FILE","Is the disk full or write protected?");
		return;
	}

	mode = FILEID;
	write(file,&mode,2);            /* Write file ID						*/

	if ( scr.fontlen <= 256 )
	{
		mode = 0;                   	/* Mode: 256 ch monofont			*/
		write(file,&mode,1);
		write(file,&scr.fontsize,1);	/* Write character height			*/

		for ( ch = scr.firstch ; ch < scr.firstch+256 ; ch++ )
		{
			write(file,char_data[ch],scr.fontsize);
		}

		close(file);
	}
	else
	{
		mode = 1;						/* Mode: 512 ch monofont			*/
		write(file,&mode,1);
		write(file,&scr.fontsize,1);	/* Write character height			*/

		for ( ch = 0 ; ch < 512 ; ch++ )
		{
			write(file,char_data[ch],scr.fontsize);
		}

		close(file);
	}

	scr.modif[scr.fontno] = FALSE;

	fqfn(filspec,scr.filespec[scr.fontno]);
	strxupr(scr.filespec[scr.fontno]);

	scr.filename[scr.fontno] = strrchr(scr.filespec[scr.fontno],'\\') + 1;
}



void getromfont(int alvalue)
{
	int ok;

	if ( scr.modif[scr.fontno] )
	{
		ok = errormsg(ERR_WARN,"YOU ARE ABOUT TO OVERWRITE AN UNSAVED FONT","Select Cancel to save your font before destroying it.");
	}
	else
		ok = TRUE;

	if ( ok )
	{
		endgraph();

		_BL = scr.fontno;
		_AL = alvalue;
		_AH = 0x11;
		geninterrupt(0x10);

		capturefont(char_data[scr.firstch],scr.firstch << 6);

		begingraph();

		scr.filename[scr.fontno] = NULL;
		scr.modif[scr.fontno] = FALSE;
	}
	else
		refreshscreen();
}



void font_command(int cmd, int x)
/* Command from the vertical all-font command bar							*/
{
	switch(cmd)
	{
		case 0:                                     /* Command: Shrink		*/
				if ( scr.fontsize > 2 )
				{
					scr.fontsize--;
					refreshscreen();
				}
				else
				{
					errormsg(ERR_DISAB,"MINIMUM FONT SIZE ALREADY SELECTED","The smallest possible font size is 2 points.");
					refreshscreen();
				}
				break;

		case 1:										/* Command: Expand		*/
				if ( scr.fontsize < 32 )
				{
					scr.fontsize++;
                    if ( scr.fontlen < 512 )
                    	scr.charfirst = scr.cie & 0xFF00;
					refreshscreen();
				}
				else
				{
					errormsg(ERR_DISAB,"MAXIMUM FONT SIZE ALREADY SELECTED","The largest possible font size is 32 points.");
					refreshscreen();
				}
				break;

		case 2:										/* Command: Save		*/
				{
					char filesp[64];

					invertmenubox(FTME_X+140,FTME_Y);

					if ( getfilename(filesp,"SAVE FONT") )
					{
						savefont(filesp);
					}

					refreshscreen();
				}
				break;

		case 3:										/* Command: Restore		*/
				{
					char filesp[64];

					invertmenubox(FTME_X+210,FTME_Y);

					if ( getfilename(filesp,"RETRIEVE FONT") )
					{
						getfont(filesp);
					}

					refreshscreen();
				}
				break;

		case 4:										/* Command: Font 0		*/
		case 5:										/* Command: Font 1		*/
				if ( scr.fontlen < 512 )
				{
					scr.firstch = scr.charfirst = ( cmd == 5 ) ? 256 : 0;
				}
				else
				{
					if ( x )
					{
						if ( bound(x,FTME_X+339,FTME_X+339+9*9) )
						{
							scr.charfirst = ( ( x - (FTME_X+339) ) / 9 ) << 5;
						}
						else
							break;
					}
					else
					{
						scr.charfirst += ( cmd == 5 ) ? 32 : -32;
					}
				}

				refreshscreen();
				break;

		case 6:										/* Command: 256 ch		*/
				scr.fontlen = 256;
				scr.firstch = scr.charfirst = scr.cie & 0xFF00;
					/* scr.charfirst will be automatically adjusted */
                    /* by refreshscreen().							*/
				refreshscreen();
				break;

		case 7:										/* Command: 512 ch		*/
				scr.fontlen = 512;
                scr.firstch = 0;
				refreshscreen();
				break;

		case 8:										/* Command: Group mode	*/
				if ( *RegID )
				{
					scr.group = !scr.group;
					invertmenubox(FTME_X+560,FTME_Y);

					errormsg(ERR_NOTDONEYET,"GROUP MODE NOT IMPLEMENTED",
                    	"This is an incomplete version of CHAREDIT.");
					refreshscreen();
				}
				else
				{
					errormsg(ERR_DISAB,"ADVANCED FEATURE - NOT ENABLED",
						"This function is only available on registered copies.");
					refreshscreen();
				}
				break;

		case 9:										/* Command: Mouse...	*/
		case 10:
		case 11:
				quickmouse(11-cmd);
				break;

		case 12:									/* Command: Copy ROM	*/
				if ( scr.fontlen == 256 )
				{
					switch ( scr.fontsize )
					{
						case 8 :
									getromfont(0x12); /* AL reg in Video 11	*/
									break;
						case 14 :
									getromfont(0x11);
									break;
						case 16 :
									if ( scr.vga )
									{
										getromfont(0x14);
										break;
									}
						default :
									errormsg(ERR_DISAB,"CURRENT FONT SIZE NOT PRESENT IN ROM","Only 8 and 14 points (and 16 for VGA only) are ROMmed");
									break;
					}
				}
				else
				{
					errormsg(ERR_DISAB,"CANNOT COPY ROM IN 512 CHARACTER MODE","All ROM fonts are 256-character fonts");
				}

				refreshscreen();
				break;


		case 13:									/* Command: Test		*/
				{
					int old_use_font = scr.usefont;

					scr.usefont = TRUE;

					endgraph();
					puttext(1,1,80,MIN(_video.screenheight,25),TESTIMAGE);

					gotoxy(1,_video.screenheight);
					textattr(LIGHTGRAY);
					cprintf("Press any key or mouse button to return to the editor... ");

					mouselimit(0,0,639,_video.screenheight*8-1);
					buttonwait();

					scr.usefont = old_use_font;

					begingraph();
				}
				break;


		case 14:									/* Command: Use Font	*/
				scr.usefont = !scr.usefont;
				invertmenubox(FTME_X+350,FTME_Y+14);
				break;

		case 15:									/* Command: Info		*/
				{
					char info2[61], info3[61];

					if ( *RegID )
					{
						sprintf( info2, "Registered to %s with ID %s", RegUser, RegID );
					}
					else
					{
						strcpy( info2, "Not registered (please do so)." );
					}

					sprintf(info3,"Copyright (C) " COPYYEAR " H. Peter Anvin * %luK free",farcoreleft()/1024);
					errormsg(ERR_INFO,"CHAREDIT version " VERSION " compiled on " __DATE__, info2, info3);
					refreshscreen();
				}
				break;

		case 16: 									/* Command: Shell		*/
				{
					char *comspec;
					int err;

					endgraph();

					textattr(LIGHTGREEN);
					printf("Use EXIT command to return to CHAREDIT\n");
					textattr(LIGHTGRAY);

					comspec = getenv("COMSPEC");
					comspec = ( comspec == NULL ) ? "\COMMAND.COM" : comspec;

					err = spawnlp(P_WAIT,comspec,comspec,NULL);

					begingraph();

					if ( err < 0 )
					{
						errormsg(ERR_EXT,"CANNOT LOAD COMMAND PROCESSOR","Not enough memory or COMSPEC incorrect.");
						refreshscreen();
					}
					else if ( err > 0 )
					{
						errormsg(ERR_EXT,"ABNORMAL SHELL TERMINATION","Your shell returned a nonzero error code.");
						refreshscreen();
					}
				}
				break;

		case 17:									/* Command: Exit		*/
        		if ( scr.modif[0] || (scr.modif[1] && scr.fontlen < 512 ) )
                {
                	scr.inedit = !errormsg(ERR_WARN,"YOUR DATA HAS NOT BEEN SAVED","Select Cancel to save your data before exiting.");
                }
                else
                	scr.inedit = FALSE;
				break;

	}
}


void selectfontsize(void)
/* Expand the font information box and select size on the 2-32 scale		*/
{
	int i, x;
	int mc, mx, my;
	char size[3];

	setcolor(DELIMITER);
	setfillstyle(SOLID_FILL,FINF_BG);
	bar3d(FINF_XEXP,FINF_Y,FINF_XEXP+620,FINF_Y+30,0,0);

	line(FINF_XEXP,FINF_Y+20,FINF_XEXP+620,FINF_Y+20);

	for ( x = FINF_XEXP+20 ; x < FINF_XEXP+620 ; x += 20 )
	{
		line(x,FINF_Y+20,x,FINF_Y+30);
	}

	setcolor(MENU_FG);
	settextjustify(CENTER_TEXT,TOP_TEXT);

	x = FINF_XEXP+10;
	for ( i = 2 ; i <= 32 ; i++ )
	{
		sprintf(size,"%d",i);
		outtextxy(x,FINF_Y+22,size);
		x += 20;
	}

	outtextxy(FINF_XEXP+310,FINF_Y+7,"Please select new font size");

	setwritemode(1);						/* XOR write mode				*/
	setcolor(MENU_FG^FINF_BG);            	/* Invert current-size box		*/

	x = FINF_XEXP - 39 + scr.fontsize * 20;	/* -39 = -2*20 + 1				*/

	for ( i = FINF_Y+21 ; i < FINF_Y+30 ; i++ )
	{
		line(x,i,x+18,i);
	}

	setwritemode(0);						/* Overwrite mode				*/

	mouseon();

	do
	{
		do
		{
			mousepress(0,&mc,&mx,&my);
		}
		while( mc == 0 );
	}
	while( !bound(my,FINF_Y+21,FINF_Y+29) || !bound(mx,FINF_XEXP+1,FINF_XEXP+619) || getpixel(mx,my) == DELIMITER );

	mouseoff();

	scr.fontsize = (mx - FINF_XEXP)/20 + 2;

    if ( scr.fontlen < 512 )
		scr.charfirst = scr.cie & 0xFF00;

	refreshscreen();
}



void setdot(int x, int y, int set)
/* Change the character-in-edit dot (x,y)									*/
{
	int xc, yc;
	char *data;

	registerstep(UND_BITEDIT);

	data = &char_data[scr.cie][y];

	*data &= ~(0x80 >> x);
	*data |= (set << (7-x));

	if ( scr.ciex )
		putpixel( scr.ciex+x, scr.ciey+y, set ? CHAR_CIE : CHAR_BG );

	putpixel( MONT_X+107-FONT_WID+x, MONT_Y+13+y, set ? MONT_FG : MONT_BG );

	setfillstyle( SOLID_FILL, set ? GRID_SQON : GRID_SQOFF );
	xc = scr.x0 + x*scr.step;
	yc = scr.y0 + y*scr.step;

	bar( xc+1, yc+1, xc+scr.step-1, yc+scr.step-1 );
}


void edit(void)
{
	int x, y, c, x1, y1, c1, i;

	mouseon();

	scr.inedit = TRUE;

	while( scr.inedit )
	{
		if ( bioskey(1) != 0 )
		{
			i = inkey();

			if ( i == KEY_NUL )
				i = 0;

			mouseoff();

			if ( i >= 0 )
			{
				selectchar(i+scr.firstch);
			}
			else if ( bound(i,KEY_F1,KEY_SF6) )
			{
				character_command(i-KEY_F1);
			}
			else if ( bound(i,KEY_CF1,KEY_CF9) )
			{
				font_command(i-KEY_CF1,0);
			}
			else if ( bound(i,KEY_AF1,KEY_AF9) )
			{
				font_command(i-KEY_AF1+9,0);
			}
			else if ( i == KEY_CHOME )
			{
				endgraph();		/* Reset the screen */
				begingraph();
			}

			mouseon();
		}

		mousepress(0,&c,&x,&y);

		if ( c )
		{
			if ( bound(x,scr.x0,scr.x1) && bound(y,scr.y0,scr.y1) )
			/* Mouse click in character grid								*/
			{
				mouseoff();

				c = getpixel(x,y);

				if ( c != GRID_COL )
				{
					c1 = ( c == GRID_SQOFF );
					setdot((x-scr.x0)/scr.step,(y-scr.y0)/scr.step,c1);

					mouseon();

					x1 = x; y1 = y;

					for ( i = 1 ; i < 256 && ( mouse(&x,&y) & MOU_LEFT ) == 0 ; i++ );

					while( mouse(&x,&y) & MOU_LEFT )
					{
						if ( x1 != x || y1 != y )
						{
							if ( bound(x,scr.x0,scr.x1) && bound(y,scr.y0,scr.y1) )
							{
								mouseoff();

								if ( c == getpixel(x,y) )
								{
									setdot(
										(x-scr.x0)/scr.step,
										(y-scr.y0)/scr.step, c1 );
								}

								mouseon();
                            }

							x1 = x; y1 = y;
						}
					}

					mousepress(0,NULL,NULL,NULL);
				}
				else
					mouseon();
			}
			else if ( bound(x,CHAR_X+4,CHAR_X1-4) && bound(y,CHAR_Y+2,scr.char_y1-2) )
			/* Mouse click in font map										*/
			{
				c = scr.charfirst + 32 * ( (y - CHAR_Y - 2)/(scr.fontsize+4) ) + (x - CHAR_X - 4)/FONT_WID;

				for ( i = 1 ; i < 256 && ( mouse(&x,&y) & MOU_LEFT ) == 0 ; i++ );
				while( mouse(&x,&y) & MOU_LEFT );

				mouseoff();

				if ( bound(x,scr.x0,scr.x1) && bound(y,scr.y0,scr.y1) )
				/* Drag to character grid									*/
				{
					registerstep(UND_ALWAYS);

					memcpy(char_data[scr.cie],char_data[c],32);
					selectchar(scr.cie);
				}
				else
				{
                	if ( bound(c,scr.firstch,scr.firstch+scr.fontlen-1) )
                    {
						selectchar(c);
                    }
                    else
                    {
						scr.cie = c;
                        scr.fontno = c/scr.fontlen;
                        scr.charfirst = scr.firstch = scr.fontno*scr.fontlen;
						refreshscreen();
                    }
				}

				mouseon();
			}
			else if ( bound(x,TEMP_X+4,TEMP_X1-4) && bound(y,TEMP_Y+2,scr.temp_y1-2) )
			/* Mouse click in temporary font map							*/
			{
				c = TEMP_FIRST + 16*((y-TEMP_Y-2)/(scr.fontsize+4)) + (x-TEMP_X-4)/FONT_WID;

				for ( i = 1 ; i < 256 && ( mouse(&x,&y) & MOU_LEFT ) == 0 ; i++ );
				while( mouse(&x,&y) & MOU_LEFT );

				mouseoff();

				if ( bound(x,scr.x0,scr.x1) && bound(y,scr.y0,scr.y1) )
				/* Drag to character grid									*/
				{
					registerstep(UND_ALWAYS);

					memcpy(char_data[scr.cie],char_data[c],32);
					selectchar(scr.cie);
				}
				else
				{
                	selectchar(c);
				}

				mouseon();
			}
			else if ( bound(x,MONT_X,MONT_X+112) && bound(y,scr.org_y0,scr.org_y1) )
			/* Mouse click in Original box									*/
			{
				registerstep(UND_ORIG);

				memcpy(char_data[scr.cie],save_cie,32);
				mouseoff();
				selectchar(scr.cie);
				mouseon();
			}
			else if ( bound(x,CHME_X,CHME_X+70) && bound(y,CHME_Y,CHME_Y+16*14) )
			/* Mouse click in vertical char-in-edit command bar				*/
			{
				mouseoff();

				if ( getpixel(x,y) != DELIMITER )	/* Borderline = ???		*/
				{
					c = ( y - CHME_Y )/14;	/* What command did we select?	*/
					character_command(c);	/* Execute the command			*/
				}

				mouseon();
			}
			else if ( bound(x,FTME_X,FTME_X+630) && bound(y,FTME_Y,FTME_Y+28) )
			/* Mouse click in horizontal all-font command bar				*/
			{
				mouseoff();

				if ( getpixel(x,y) != DELIMITER )	/* Borderline = ???		*/
				{
					c = ( x - FTME_X ) / 70 + 9 * ( y >= FTME_Y+14 );
													/* Command number		*/

					font_command(c,x);
				}

				mouseon();
			}
			else if ( bound(x,FINF_X,FINF_X+112) && bound(y,FINF_Y,FINF_Y+30) )
			/* Mouse click in font information box (set size command)		*/
			{
				mouseoff();
				selectfontsize();
				mouseon();
			}
		}
	}

	mouseoff();
}


main(int argc,char *argv[])
{
	char *swset;
	int i, non_reg;

    clrscr();
    textattr(LIGHTGRAY);

	cprintf("\nCharEdit version " VERSION " compiled on " __DATE__ "\r\n"
			"Copyright (C) " COPYYEAR " H. Peter Anvin - all rights reserved\r\n\n");

	swset =	seekswitch(&argc,argv);

	scr.setpal = ( strchr(swset,'N') == NULL );

	save_data = farmalloc((512+TEMP_CNT)*32);
    										/* Get save_data far memory		*/

	if ( save_data == NULL )
	{
    	textattr(WHITE+BLINK);
		cprintf("ERROR: Insufficient memory available.\r\n\n");
        textattr(LIGHTGRAY);
		exit(2);
	}

	non_reg = FALSE;

	if ( resetmouse() == 0 )
	{
		cprintf("\nERROR: No mouse attached, or mouse driver not loaded\n");
		farfree(save_data);
		return(6);
	}

	if ( *RegID )
	{
		cprintf("This copy has ID %s and is registered to %s\r\n\n", RegID, RegUser );
	}
	else
	{
		textattr(WHITE+BLINK);
		cprintf("Unregistered copy - please support shareware!\r\n\n");
		delay(3000);
	}
	textattr(LIGHTGRAY);

	scr.fontsize = capturefont(char_data[0],0);
	capturefont(char_data[256],0x4000);

    for ( i = 0 ; i < TEMP_CNT ; i++ )
    {
    	memcpy(char_data[TEMP_FIRST+i],char_data[temp_init[i]],32);
    }

	scr.fontlen = 256;
	scr.firstch = scr.charfirst = 0;
	farmemcpy(save_data,char_data,32*512);
	farmemset(&save_data[TEMP_FIRST],0,32*TEMP_CNT);

	scr.cie = scr.firstch+' ';	/* Default start-of-edit character			*/
	scr.filename[0] = scr.filename[1] =  NULL;		/* No filename given	*/
	scr.use32 = scr.group = scr.modif[0] = scr.modif[1] = FALSE;
	scr.usefont = TRUE;

	undo.cnt = undo.recnt = 0;
	undo.free = undo.size = ( coreleft() - SAVECORE ) / sizeof(struct undo_item);
	undo.ptr = undo.block = malloc( undo.free * sizeof(struct undo_item) );
	undo.endblock = undo.block + undo.size;

	scr.mouspeed = 1;

	if ( begingraph() < 0 )
	{
    	textattr(WHITE+BLINK);
		cprintf("ERROR: No EGA/VGA loaded, monochrome monitor,\r\nor EGA memory less than 256K\r\n");
		free(undo.block);
		farfree(save_data);
		return(6);
	}

	if ( argc >= 2 && non_reg )
	{
		getfont(expandedname(argv[1]));
		refreshscreen();
	}

	edit();

	endgraph();

	puttext(1,1,80,MIN(_video.screenheight,25),TESTIMAGE);

	gotoxy(1,_video.screenheight - 1);

	free(undo.block);
	farfree(save_data);

	return(0);
}
