;
; EGAPUTCH.ASM  -- C routine to display a character on-screen in graphic
;                  mode 10h. Since the BGI cannot define rastered characters
;                  other than its own 8x8, we have to do it ourselves.
;                  Only for use with 640x200, 640x350 and 640x480 native
;                  EGA/VGA modes.
;
; From The Programmer's Guide to PC and PS/2 Video Systems, page 289, modified.
;
; Small memory model.
; void egaputch(int ch, int x, int y, int page, int fg, int bg,
;               void *font, int size);
; ch    = character
; x,y   = coordinates
; page  = screen page
; fg,bg = foreground/background colours
; font  = pointer to 32-bytes-per-char font table
; size  = height of font in pixels
;

BytesPerLine    equ 80                          ; 80 bytes/pixel line
RMWbits         equ 00h                         ; Read-Modify-Write bits

_TEXT           segment byte public 'CODE'
_TEXT           ends

DGROUP          group   _DATA,_BSS

_DATA           segment word public 'DATA'
d@              label   byte
d@w             label   word
OriginOffset    dw 0                            ; Font page base
_DATA           ends

_BSS            segment word public 'BSS'
b@              label   byte
b@w             label   word
_BSS            ends

_TEXT           segment byte public 'CODE'

                assume  cs:_TEXT,ds:DGROUP,ss:DGROUP

                public _egaputch

; PixelAddr10: From The Programmer Guide to PC & PS/2 Video Systems, page 96.

PixelAddr10     proc near
                mov cl,bl

                push dx
                mov dx,BytesPerLine
                mul dx
                pop dx

                shr bx,1
                shr bx,1
                shr bx,1
                add bx,ax
                add bx,OriginOffset

                mov ax,0A000h                   ; Video Buffer Segment
                mov es,ax

                and cl,7
                xor cl,7
                mov ah,1
                ret
PixelAddr10     endp

arg_ch          equ word ptr [bp+4]             ; Stack frame: argument list
arg_x           equ word ptr [bp+6]
arg_y           equ word ptr [bp+8]
arg_page        equ word ptr [bp+10]
arg_fg          equ byte ptr [bp+12]
arg_bg          equ byte ptr [bp+14]
arg_font        equ word ptr [bp+16]
arg_size        equ word ptr [bp+18]
var_shift       equ [bp-2]                      ; Local variable

_egaputch       proc near                       ; proc far in large-code models
                push bp
                mov bp,sp
                push ax                         ; Reserve space for var_shift
                push di
                push si
                push ds

                mov ax,arg_page                 ; Page number
                mov cl,15                       ; Shift 15 bits = 32K/page
                shl ax,cl
                mov OriginOffset,ax             ; Start of page

                mov ax,arg_y
                mov bx,arg_x
                call PixelAddr10

                inc cx
                and cl,7
                mov ch,0FFh
                shl ch,cl
                mov var_shift,cx

                push es                         ; Save video buffer segment

                mov cx,arg_size                 ; Load size of font in points
                mov di,arg_ch                   ; Load character
                shl di,1                        ; Shift left 5 steps
                shl di,1
                shl di,1
                shl di,1
                shl di,1
                add di,arg_font                 ; Add initial address
                push ds
                pop es                          ; ES:DI -> char definitions

                pop ds
                mov si,bx                       ; DS:SI -> video buffer

                mov dx,3CEh
                mov ax,0A05h
                out dx,ax
                mov ax,(RMWbits shl 8) + 3
                out dx,ax
                mov ax,0007h
                out dx,ax

                mov bl,arg_fg
                mov bh,arg_bg
                cmp byte ptr var_shift,0
                jne L20

                mov al,8                        ; For aligned x-coordinates
L10:            mov ah,es:[di]
                out dx,ax
                and [si],bl

                not ah
                out dx,ax
                and [si],bh

                inc di
                add si,BytesPerLine
                loop L10
                jmp short LExit

L20:            push cx                         ; For nonaligned x-coordinates
                mov cx,var_shift

                mov al,es:[di]
                xor ah,ah
                shl ax,cl

                push ax

                mov al,8
                out dx,ax

                and [si],bl

                not ch
                xor ah,ch
                out dx,ax

                and [si],bh

                pop ax

                mov ah,al
                mov al,8
                out dx,ax

                inc si

                and [si],bl

                not ch
                xor ah,ch
                out dx,ax

                and [si],bh

                inc di
                dec si
                add si,BytesPerLine

                pop cx
                loop L20

LExit:          mov ax,0FF08h
                out dx,ax
                mov ax,0005h
                out dx,ax
                mov ax,0003h
                out dx,ax
                mov ax,0F07h
                out dx,ax

                pop ds
                pop si
                pop di
                mov sp,bp
                pop bp
                ret
_egaputch       endp

_TEXT           ends

_DATA           segment word public 'DATA'
s@              label   byte
_DATA           ends

                end
