/*																			*/
/*	RTCTIME.C	-- Instructions to read the Real Time Clock					*/
/*																			*/
/*	realtime() returns 0 if values read; 1 if error							*/
/*																			*/

#include <stdlib.h>
#include <dos.h>
#include <pushpop.h>

#define CARRY	(_FLAGS & 0x01)

typedef union
{
	unsigned char byte;

	struct
	{
		unsigned one : 4;
		unsigned ten : 4;
	}
	d;
}
bcdpair;


int realtime(int *year, int *month, int *day, int *hour, int *min, int *sec)
{
	bcdpair cy, yr, mo, dy, hr, mi, se;

	_AH = 0x04;

	geninterrupt(0x1A);

	cy.byte = _CH;
	yr.byte = _CL;
	mo.byte = _DH;
	dy.byte = _DL;

	if ( CARRY )
		return(1);

	_AH = 0x02;

	geninterrupt(0x1A);

	hr.byte = _CH;
	mi.byte = _CL;
	se.byte = _DH;

	if ( CARRY )
		return(1);

	*year = cy.d.ten * 1000 + cy.d.one * 100 + yr.d.ten * 10 + yr.d.one;
	*month = mo.d.ten * 10 + mo.d.one;
	*day = dy.d.ten * 10 + dy.d.one;
	*hour = hr.d.ten * 10 + hr.d.one;
	*min = mi.d.ten * 10 + mi.d.one;
	*sec = se.d.ten * 10 + se.d.one;

	return(0);
}



int updateclock(void)
{
	int yr, mo, dy, hr, mi, se, se1;

	if ( realtime(&yr,&mo,&dy,&hr,&mi,&se1) )
		return(1);

	do
	{
		realtime(&yr,&mo,&dy,&hr,&mi,&se);
	}
	while ( se == se1 );					/* Wait for start of new second	*/

	_CH = hr;
	_CL = mi;
	_DH = se;
    _DL = 0;
	_AH = 0x2D;
	geninterrupt(0x21);

	if ( _AL )
		return(1);

	_CX = yr;
	_CH = mo;
	_DL = dy;
	geninterrupt(0x21);

	if ( _AL )
		return(1);

	return(0);
}


