/* WINDOW01.C contains window-handling functions for Turbo C, part 1 */

#include <dos.h>
#include <stdio.h>
#include <stdlib.h>
#include <conio.h>

enum WSTAT { W_OK, W_NOTINIT, W_NOSPACE, W_EMPTY, W_RANGE, W_ILLEGAL };
enum WMODE { W_NORMAL, W_SPECIAL };
enum WSTYLE { W_SINGLE, W_DOUBLE, W_DVERT, W_DHORIZ, W_NOBORDER = 64,
    W_SHADOW = 128 };


#define zerowindows() while( killwindow() == W_OK )

enum WSTAT disablewindows(void);
enum WSTAT killwindow(void);


int WindowColor = GREEN;
enum WSTYLE WindowStyle = W_SINGLE;

struct WindowSysStruct
{
	unsigned *storage;
	unsigned *bufptr;
	int wsize;
	int wlim;
	int spaceleft;
    int maxcol;
    int maxline;
    int shadowwid;
    int colour;
};


struct WindowSysStruct _WindowSys = { NULL, NULL, 0, 0, 0, 80, 25, 2, 1 };

#define order(a,b) if (a>b) { dummy = a; a=b; b=dummy; }
#define TESTINIT if ( _WindowSys.buf == NULL ) return(W_NOTINIT)
#define xwindows() ( _WindowSys.spaceleft < _WindowSys.wsize )

#define WS          _WindowSys


enum WSTAT initwindows(int size, int limit)
/* Set screen mode and allocate (size) words of memory for ...window()
   functions. (limit) words can only be accessed with W_SPECIAL set. If
   ( limit>size ), set ( limit = size ).

   Current number of useful lines taken from current gettextinfo(). DO
   NOT CHANGE VIDEO MODES WHILE THE WINDOW SYSTEM IS OPERATIVE.

   Exit codes:
   W_OK			Everything OK
   W_NOSPACE	malloc() failed
																		   */
{
	struct text_info coninfo;

    if ( WS.bufptr != NULL )
		disablewindows();

	_AH = 15;								/* Check current mode, or EGA */
	geninterrupt(0x10);
    WS.colour = (_AL != 7);

	gettextinfo(&coninfo);

    WS.maxline = coninfo.screenheight;
    WS.maxcol = coninfo.screenwidth;

    WS.shadowwid = ( 7 * WS.maxcol ) / ( 8 * WS.maxline );

    window(1,1,WS.maxcol,WS.maxline);
	clrscr();

    if ( (WS.storage = (unsigned *) malloc( size * sizeof(int) ) ) == NULL )
		return(W_NOSPACE);

    WS.bufptr = WS.storage;
    WS.wsize = size;
    WS.wlim = ( size<limit ) ? size : limit;
    WS.spaceleft = size;

	return(W_OK);
}



enum WSTAT disablewindows(void)
/* Free the window buffer, and disable the window system.					*/
/* For use of the window system after disabling, a new call to 				*/
/* initwindows() will be needed.											*/
/*																			*/
/* Return values:															*/
/*		W_OK		... system freed OK										*/
/*		W_NOTINIT	... system never initialized							*/
{
    if ( WS.bufptr == NULL )
	{
        return(W_NOTINIT);
	}
	else
	{
		zerowindows();
        free(WS.storage);
        WS.bufptr = WS.storage = NULL;
		return(W_OK);
	}
}
